﻿using Kairai;

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Numerics;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Threading;
using System.Threading.Tasks;
using Windows.ApplicationModel;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.Graphics.Display;
using Windows.Storage;
using Windows.Storage.Streams;
using Windows.Storage.Pickers;
using Windows.System.Threading;
using Windows.UI.Core;
using Windows.UI.Popups;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;

namespace KumaApp
{
    public enum GameStatus
    {
        Initialize,
        Load,
        Wait,
        Ready,
        Fall,
        Fight,
        GameOver,
        Restart,
        Goal,
        Max
    };

    public enum Mesh
    {
        Kuma = 0,
        Corn,
        DoraNeko,
        ShiroNeko,
        Grass,
        Ball,
        Goal,
        Max
    };

    public enum Action
    {
        Walk = 0,
        Attack,
        Hitted,
        Max
    };

    public struct BallInf
    {
        public float count;
        public float dx;
        public float dz;

        public BallInf(float _cnt)
        {
            count = _cnt;
            dx = 0;
            dz = 0;
        }
    };

    public class Util
    {
        MetroMesh m_kuma;
        GameStatus m_status;
        KairaiApi m_kairai;
        Point m_mousePos;
        Rect m_ground = new Rect(-48, -56, 64, 64);
        float m_fall;
        BallInf m_ball;
        int m_maxTiles = 4;
        float[,] m_grassPos =
            {
                {0,-8},
                {0,-40},
                {-32,-8},
                {-32,-40}
            };
        int m_maxCorns = 5;
        float[,] m_cornPos =
            {
                {-16,-8},
                {-0,-40},
                {-24,-16},
                {-32,-8},
                {-16,-40}
            };

        public GameStatus Status
        {
            get { return m_status; }
            set { m_status = value; }
        }
        public Point MousePos
        {
            get { return m_mousePos; }
            set { m_mousePos = value; }
        }

        void Normalize(ref float x, ref float z)
        {
            float len = (float)Math.Sqrt(x * x + z * z);
            if(len > 0)
            {
                x /= len;
                z /= len;
            }
        }

        public Util(KairaiApi kai)
        {
            m_status = GameStatus.Initialize;
            m_kairai = kai;
        }
        
        int GetKuma()
        {
            int kuma = (int)Mesh.Kuma;
            m_kuma = m_kairai.GetMesh(kuma);
            return kuma;
        }

        void SetKuma()
        {
            int kuma = (int)Mesh.Kuma;
            m_kairai.SetMesh(kuma, m_kuma);
        }

        public bool Ready()
        {
            return (m_status == GameStatus.Ready);
        }

        void MoveMesh(Mesh meshEnum, float x, float y, float z)
        {
            int meshNo = (int)meshEnum;
            MetroMesh mesh = m_kairai.GetMesh(meshNo);
            mesh.OffsetX = x;
            mesh.OffsetY = y;
            mesh.OffsetZ = z;
            m_kairai.SetMesh(meshNo, mesh);
        }

        void StartWalking(Mesh meshEnum, float x, float z)
        {
            MoveMesh(meshEnum, x, 0, z);
            m_kairai.PlayMesh((int)meshEnum, (int)Action.Walk, true, true);
        }

        void StartGame()
        {
            //Game
            m_kairai.SetGame(true);
            m_mousePos.X = 0;
            m_mousePos.Y = 0;

            //Ball
            m_ball = new BallInf(0);
            MoveMesh(Mesh.Ball, -8, 0, 0);

            //Goal
            m_kairai.SetRadius((int)Mesh.Goal, 1);
            MoveMesh(Mesh.Goal, -60, 0, -48);

            //Kumamon
            MoveMesh(Mesh.Kuma, 0, 0, 0);
            StartWalking(Mesh.Kuma, 0, 0);
            MoveKuma(new Point(0, 0));

            //Cats
            StartWalking(Mesh.DoraNeko, -48, -48);
            StartWalking(Mesh.ShiroNeko, 8, -32);

            //Ground
            int grassNo = (int)Mesh.Grass;
            m_kairai.SetPass(grassNo, true);
            m_kairai.SetMaxCopies(grassNo, m_maxTiles-1);
            for (int i = 0; i < m_maxTiles; i++)
            {
                float x = m_grassPos[i, 0];
                float y = 0;
                float z = m_grassPos[i, 1];

                if(i < 1)
                {
                    MoveMesh(Mesh.Grass, x, 0, z);
                }
                else
                {
                    m_kairai.SetCopyMove(grassNo, i - 1, x, y, z, 0);
                }
            }

            //Corn
            int cornNo = (int)Mesh.Corn;
            m_kairai.SetMaxCopies(cornNo, m_maxCorns - 1);
            for (int i = 0; i < m_maxCorns; i++)
            {
                float x = m_cornPos[i, 0];
                float y = 0;
                float z = m_cornPos[i, 1];

                if (i < 1)
                {
                    MoveMesh(Mesh.Corn, x, 0, z);
                }
                else
                {
                    m_kairai.SetCopyMove(cornNo, i - 1, x, y, z, 1);
                }
            }
        }

        public async void LoadMeshes()
        {
            if (m_status == GameStatus.Load)
            {
                return;
            }
            m_status = GameStatus.Load;

            String fileName = "KumaGame.xbone";
            StorageFolder localFolder = Package.Current.InstalledLocation;
            var loadFile = await localFolder.GetFileAsync(fileName);
            var randomStream = await loadFile.OpenAsync(FileAccessMode.Read);

            ulong buffSize = randomStream.Size;
            byte[] buffer = new byte[buffSize];

            DataReader reader = new DataReader(randomStream.GetInputStreamAt(0));
            await reader.LoadAsync((uint)buffSize);
            reader.ReadBytes(buffer);

            m_kairai.LoadBin(buffer);
            StartGame();

            m_status = GameStatus.Wait;
            m_kairai.SetKairaiState(KAIRAI_STATE.KAIRAI_READY);
        }

        public void MoveKuma(Point pos)
        {
            m_mousePos = pos;
            if (!Ready())
            {
                return;
            }
            int kuma = GetKuma();
            if (m_kairai.PlayingAction(kuma) != (int)Action.Walk)
            {
                return;
            }

            // Move Kumamon toward the mouse position.
            pos = m_kairai.GetDirection(kuma, pos);
            float x = (float)pos.X;
            float z = (float)pos.Y;
            m_kairai.MoveMesh(kuma, x, 0, z);

            // Move cats toward the position of Kumamon.
            x = m_kuma.OffsetX;
            z = m_kuma.OffsetZ;
            m_kairai.MoveMesh((int)Mesh.DoraNeko, x, 0, z - 3);
            m_kairai.MoveMesh((int)Mesh.ShiroNeko, x, 0, z - 2);
        }

        public void Callback()
        {
            int kuma = GetKuma();
            int enemy = m_kairai.Collided(kuma);

            Fall();
            Fight(enemy);
            Ball(enemy);
        }

        void Fall()
        {
            float kx = m_kuma.OffsetX;
            float kz = m_kuma.OffsetZ;
            switch (m_status)
            {
                case GameStatus.Ready:
                    if (kx >= m_ground.Left &&
                        kx <= m_ground.Right &&
                        kz >= m_ground.Top &&
                        kz <= m_ground.Bottom)
                    {
                        return;
                    }
                    m_fall = 0;
                    m_status = GameStatus.Fall;
                    break;

                case GameStatus.Fall:
                    MoveMesh(Mesh.Kuma, kx, m_fall, kz);
                    m_fall -= 0.5F;
                    if( m_fall < -40)
                    {
                        m_status = GameStatus.GameOver;
                    }
                    break;

                case GameStatus.GameOver:
                    m_fall -= 0.5F;
                    if (m_fall < -140)
                    {
                        m_status = GameStatus.Restart;
                    }
                    break;

                case GameStatus.Restart:
                    m_fall -= 0.5F;
                    if (m_fall < -240)
                    {
                        StartGame();
                        m_status = GameStatus.Ready;
                    }
                    break;

                default:
                    break;
            }
        }

        void Ball(int enemy)
        {
            int ball = (int)Mesh.Ball;
            int thing = m_kairai.Collided(ball);
            if(thing == (int)Mesh.Goal)
            {
                if(m_status != GameStatus.Goal)
                {
                    m_status = GameStatus.Goal;
                    m_kairai.MoveMesh((int)Mesh.Kuma, m_kuma.OffsetX, 0, m_kuma.OffsetZ);
                }
            }

            MetroMesh ballMesh = m_kairai.GetMesh(ball);
            m_ball.count += (float)Math.PI / 64;
            if (m_ball.count >= Math.PI * 2)
            {
                m_ball.count = 0;
            }
            m_ball.dx = m_ball.dx * 0.9F;
            m_ball.dz = m_ball.dz * 0.9F;

            ballMesh.OffsetX = ballMesh.OffsetX + m_ball.dx;
            ballMesh.OffsetY = (float)Math.Sin(m_ball.count) * 2.0F;
            ballMesh.OffsetZ = ballMesh.OffsetZ + m_ball.dz;

            switch (enemy)
            {
                case (int)Mesh.Ball:
                    break;

                default:
                    m_kairai.SetMesh(ball, ballMesh);
                    return;
            }

            //Distance
            m_ball.dx = ballMesh.OffsetX - m_kuma.OffsetX;
            m_ball.dz = ballMesh.OffsetZ - m_kuma.OffsetZ;
            Normalize(ref m_ball.dx, ref m_ball.dz);

            ballMesh.OffsetX = ballMesh.OffsetX + m_ball.dx;
            ballMesh.OffsetZ = ballMesh.OffsetZ + m_ball.dz;
            m_kairai.SetMesh(ball, ballMesh);
        }

        void Fight(int enemy)
        {
            if (!Ready())
            {
                return;
            }
            switch (enemy)
            {
                case (int)Mesh.DoraNeko:
                case (int)Mesh.ShiroNeko:
                    break;

                default:
                    return;
            }

            m_status = GameStatus.Fight;

            int kuma = (int)Mesh.Kuma;
            MetroMesh neko = m_kairai.GetMesh(enemy);

            //Distance
            float dx = neko.OffsetX-m_kuma.OffsetX;
            float dz = neko.OffsetZ-m_kuma.OffsetZ;
            Normalize(ref dx, ref dz);

            switch (m_kairai.PlayingAction(kuma))
            {
                //Kumamon attacks a cat.
                case (int)Action.Attack:
                    if (m_kairai.PlayingAction(enemy) != (int)Action.Hitted)
                    {
                        MoveMesh((Mesh)enemy,
                            neko.OffsetX + dx * 6,
                            0,
                            neko.OffsetZ + dz * 6);
                        m_kairai.PlayMesh(enemy, (int)Action.Hitted, false, false);
                    }
                    break;

                //Kumamon is attacked by a cat.
                default:
                    m_kuma.OffsetX = m_kuma.OffsetX - dx * 4;
                    m_kuma.OffsetZ = m_kuma.OffsetZ - dz * 4;
                    m_kuma.AngleY = 
                        (float)(Math.Atan2(dz, -dx) - Math.PI / 2);

                    SetKuma();
                    m_kairai.PlayMesh(kuma, (int)Action.Hitted, false, false);
                    break;
            }
            m_status = GameStatus.Ready;
        }

        public void Attack(Point pos)
        {
            int kuma = (int)Mesh.Kuma;
            if (m_kairai.PlayingAction(kuma) != (int)Action.Walk)
            {
                return;
            }

            m_kairai.PlayMesh(kuma, (int)Action.Attack, false, false);
            pos = m_kairai.GetDirection(0, pos);
            double len = Math.Sqrt(pos.X * pos.X + pos.Y * pos.Y) / 4;

            m_kuma = m_kairai.GetMesh(kuma);
            pos.X = (pos.X - m_kuma.OffsetX) / len;
            pos.Y = (pos.Y - m_kuma.OffsetY) / len;
            m_kuma.OffsetX = m_kuma.OffsetX + (float)pos.X;
            m_kuma.OffsetZ = m_kuma.OffsetZ + (float)pos.Y;
            SetKuma();
        }
    }
}
